#define PYGII_MODULE
#include "pygii.h"

/**
 ** Hack from ggiwrapper to allow multiple args to giiOpen
 **/
typedef gii_input_t _varidx(char**);
typedef _varidx*varidx;
#define v_ gii_input_t
v_ open0(char**v){return giiOpen(NULL);} 
v_ open1(char**v){return giiOpen(v[0],NULL);} 
v_ open2(char**v){return giiOpen(v[0],v[1],NULL);} 
v_ open3(char**v){return giiOpen(v[0],v[1],v[2],NULL);} 
v_ open4(char**v){return giiOpen(v[0],v[1],v[2],v[3], NULL);} 
v_ open5(char**v){return giiOpen(v[0],v[1],v[2],v[3],v[4],NULL);} 
v_ open6(char**v){return giiOpen(v[0],v[1],v[2],v[3],v[4],v[5],NULL);} 
v_ open7(char**v){return giiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],NULL);} 
v_ open8(char**v){return giiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],NULL);}
v_ open9(char**v){return giiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],NULL);}
v_ open10(char**v){return giiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],NULL);}

#define kcah(n) &open ## n
varidx _giiOpenWithArgs[] ={kcah(0),kcah(1),kcah(2),kcah(3),kcah(4),kcah(5),
			   kcah(6),kcah(7),kcah(8),kcah(9),kcah(10)};
#undef v_
#undef kcah
/**
 ** bad-hacks-never-end
 **/

/* helpers */
static inline int pygii_get_int(PyObject *o, int *i) {
	int a;
	a=PyInt_AsLong(o);
	if(PyErr_Occurred())
		return -1;
	*i=a;
	return 0;
}

static inline int pygii_get_time(PyObject *o, struct timeval * t) {
	struct timeval r;
	if(!PyArg_ParseTuple(o, "ll", &r.tv_sec, &r.tv_usec))
		return -1;
	*t=r;
	return 0;
}

/***
 *** Get info form all devices bound to a given input.
 ***/
const char inp_devices_doc[]="
Returns a description of all devices bound to this input source.
    [ i0,i1,... ] where i are, for each device:
    [ input origin,
      device origin,
      device shortname,
      device long name,
      event mask (generated events),
      number of buttons,
      number of axes ]
";

static PyObject * pygii_device_info(struct gii_input * inp) {
	struct gii_input * curr;
	gii_deviceinfo   *di;
	PyObject * ret;
	PyObject * tmp;
	
	curr=inp;
	ret=PyList_New(0);
	do {
		for(di=curr->devinfo;di;di=di->next) {
			tmp=PyList_New(7);
			PyList_SET_ITEM(tmp,0,PyInt_FromLong(curr->origin));
			PyList_SET_ITEM(tmp,1,PyInt_FromLong(di->origin));
			PyList_SET_ITEM(tmp,2,PyString_FromString(di->dev->shortname));
			PyList_SET_ITEM(tmp,3,PyString_FromString(di->dev->longname));
			PyList_SET_ITEM(tmp,4,PyInt_FromLong(di->dev->can_generate));
			PyList_SET_ITEM(tmp,5,PyInt_FromLong(di->dev->num_buttons));
			PyList_SET_ITEM(tmp,6,PyInt_FromLong(di->dev->num_axes));
			PyList_Append(ret,tmp);
		}
		curr = curr->next;
	} while(curr != inp);	/* looped through once. */
	return ret;
}

/* types */
#include "pygii_event.inc"
#include "pygii_input.inc"

/* modules */
const char gii_init_doc[]="
Initialize GII. It must be called at least one before using any other function.
Successive calls to this function has no effect but increasing the use count.

Parameter: None
Return:    None
";
static PyObject* gii_init(void) {
	if(giiInit()) {
		PyErr_SetString(PyExc_SystemError, "giiInit() failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

const char gii_exit_doc[]="
Terminate a GII session. GII is really terminated when the use count is down to 0.

Parameters: None
Return:     None
";
static PyObject* gii_exit(void) {
	if(giiExit()<0) {
		PyErr_SetString(PyExc_SystemError, "giiExit() failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

const char gii_open_doc[]="
Open a new input source with the given arguments. If no argument is given, it
tries to open the default input as set in the GII_INPUT environment variable.

Parameters: (s,...) : Up to ten strings defining the input modules
Return:     0 : A new Input object
";
static PyObject* gii_open(PyObject * self, PyObject * args) {
	return newInputObject(args);
}

static struct PyMethodDef gii_methods[] = {
	{"giiInit",  (PyCFunction)gii_init, METH_NOARGS,  (char*)gii_init_doc },
	{"giiExit",  (PyCFunction)gii_exit, METH_NOARGS,  (char*)gii_exit_doc },
	{"Input", (PyCFunction)gii_open, METH_VARARGS,  (char*)gii_open_doc },
	{"Event", (PyCFunction)EventConstruct, METH_NOARGS, (char*)event_construct_doc },
	{NULL, NULL}
};

const char gii_doc[]="
PyGII 0.1

PyGII allows python programmers to access the GII input management system
through an elegant and effective (in the author opinion) python interface.

The first thing to do is to call giiInit().
Good luck.

Author
- Eric Faurot <eric.faurot@info.unicaen.fr>
";

DL_EXPORT(void)
initgii(void)  {
	
	PyObject *m,*t,*d;
	static Pygii_C_api capi;
	
	capi.input_type=&input_type;
	capi.event_type=&event_type;
	capi.neweventobjectfromevent=&newEventObjectFromEvent;
	capi.deviceinfo=&pygii_device_info;

	m=Py_InitModule3("gii", gii_methods,(char*)gii_doc);
	t=PyCObject_FromVoidPtr(&capi, NULL);
	d=PyModule_GetDict(m);
	
        PyDict_SetItemString(d, "_gii_C_API", t);
        Py_DECREF(t);

#define intconst(x) PyDict_SetItemString(d, #x, Py_BuildValue("i", x))
#define renintconst(x,y) PyDict_SetItemString(d, #y, Py_BuildValue("i", x))
	// event types
	intconst(evNothing);
	intconst(evCommand);
	intconst(evInformation);
	intconst(evExpose);
	intconst(evKeyPress);
	intconst(evKeyRelease);
	intconst(evKeyRepeat);
	intconst(evPtrRelative);
	intconst(evPtrAbsolute);
	intconst(evPtrButtonPress);
	intconst(evPtrButtonRelease);
	intconst(evValAbsolute);
	intconst(evValRelative);
	intconst(evLast);
	
	// event mask bits
	intconst(emNothing);
	intconst(emCommand);
	intconst(emInformation);
	intconst(emExpose);
	intconst(emKeyPress);
	intconst(emKeyRelease);
	intconst(emKeyRepeat);
	intconst(emKey);
	intconst(emKeyboard);
	intconst(emPtrRelative);
	intconst(emPtrAbsolute);
	intconst(emPtrButtonPress);
	intconst(emPtrButtonRelease);
	intconst(emPtrMove);
	intconst(emPtrButton);
	intconst(emPointer);
	intconst(emValRelative);
	intconst(emValAbsolute);
	intconst(emValuator);
	intconst(emZero);
	intconst(emAll);

	// misc
	
	/* special origins */

	renintconst(GII_EV_ORIGIN_NONE, evOriginNone);
	renintconst(GII_EV_ORIGIN_SENDEVENT, evOriginSendEvent);

	/* special target */

	renintconst(GII_EV_TARGET_ALL, evTargetAll);
	renintconst(GII_EV_TARGET_QUEUE, evTargetQueue);
	
	/* cmd flags */
	
	renintconst(GII_CMDFLAG_NODATA,   cmdflagNoData);
	renintconst(GII_CMDFLAG_PRIVATE,  cmdflagPrivate);
	renintconst(GII_CMDFLAG_EXTERNAL, cmdflagExternal);

	/* command codes */
	
	renintconst(GII_CMDCODE_EVENTLOST,     cmdEventLost);
	renintconst(GII_CMDCODE_PREFER_ABSPTR, cmdPreferAbsPtr);
	renintconst(GII_CMDCODE_PREFER_RELPTR, cmdPreferRelPtr);
	renintconst(GII_CMDCODE_GETDEVINFO,    cmdGetDevInfo);
	renintconst(GII_CMDCODE_GETVALINFO,    cmdGetValInfo);

	/* buttons and axes */

	renintconst(GII_NUM_MAX,     numMax);
	renintconst(GII_NUM_UNKNOWN, numUnknown);


	/* physical quantities */
	
	renintconst(GII_PT_UNKNOWN,         ptUnknown);			/* unknown */
	renintconst(GII_PT_TIME,            ptTime);			/* base unit s */
	renintconst(GII_PT_FREQUENCY,       ptFrequency);        	/* base unit 1/s (Hz) */
	renintconst(GII_PT_LENGTH,          ptLength);			/* base unit m */
	renintconst(GII_PT_VELOCITY,        ptVelocity); 		/* base unit m/s */
	renintconst(GII_PT_ACCELERATION,    ptAcceleration); 		/* base unit m/s^2 */
	renintconst(GII_PT_ANGLE,           ptAngle);			/* base unit radian */
	renintconst(GII_PT_ANGVELOCITY,     ptAngleVelovity);		/* base unit radian/s */
	renintconst(GII_PT_ANGACCELERATION, ptAngleAcceleration);	/* base unit radian/s^2 */
	renintconst(GII_PT_AREA,            ptArea);			/* base unit m^2 */
	renintconst(GII_PT_VOLUME,          ptVolume); 			/* base unit m^3 */
	renintconst(GII_PT_MASS,            ptMass);			/* base unit kg */
	renintconst(GII_PT_FORCE,           ptForce);			/* base unit N (kg*m/s^2) */
	renintconst(GII_PT_PRESSURE,        ptPressure);		/* base unit N/m^2 (Pa) */
	renintconst(GII_PT_TORQUE,          ptTorque);	 		/* base unit Nm */
	renintconst(GII_PT_ENERGY,          ptEnergy);			/* base unit Nm, VAs, J */
	renintconst(GII_PT_POWER,           ptPower);			/* base unit Nm/s, VA, W */
	renintconst(GII_PT_TEMPERATURE,	    ptTemperature);             /* base unit K */
	renintconst(GII_PT_CURRENT,         ptCurrent);			/* base unit A */
	renintconst(GII_PT_VOLTAGE,         ptVoltage);			/* base unit V (kg*m^2/(As^3)) */
	renintconst(GII_PT_RESISTANCE,      ptResistance);		/* base unit V/A (Ohm) */
	renintconst(GII_PT_CAPACITY,        ptCapacity);		/* base unit As/V (Farad) */
	renintconst(GII_PT_INDUCTIVITY,     ptInductivity);             /* base unit Vs/A (Henry) */
	renintconst(GGI_PT_LAST,            ptLast);

	/* misc */

	renintconst(GII_PBUTTON_LEFT,   pbuttonLeft);
	renintconst(GII_PBUTTON_RIGHT,  pbuttonRight);
	renintconst(GII_PBUTTON_MIDDLE, pbuttonMiddle);

#undef intconst
#undef renintconst
}
